
// This class is used to support a continuous plot of attribute
// values that exist only at a fixed number of points in 2-space

// (we will have a bunch of particles at different locations with 
//  different densities, for every pixel in our display, we'd like
//  to plot a color cooresponding to the density of the closest 
//  particle)

import java.util.*;

public class AttributeGrid
{
    private final static float avg_cell = 1.0f;       // how many pts on average we want in a cell
                                                      // this was optimized by hand
    private final static int average_over = 5;        // how many NN to average over
                                                      // I tried 15, but it didn't make much visible difference    
    private final static float FLAG = 99999.0f;       // used by the minimum algorithm  
    public float max_x, max_y;
    private int num_cells;
    private Vector[][] cell;
    
    public AttributeGrid(int num_particles, float mx, float my)
    {
        // compute the size of our grid based on the number of particles
        // and the average number we'd like in each cell
 
        num_cells = (int) Math.sqrt((double) num_particles / (double) avg_cell);
        
        System.out.println("creating cells, num_cells = " + num_cells);
        
        cell = new Vector[num_cells][num_cells];
        
        for (int i=0; i<num_cells; i++)
            for (int j=0; j<num_cells; j++)
                cell[i][j] = new Vector();
        
        max_x = mx;
        max_y = my;
    }
    
    // given a particle, add to the approriate cell
    public void add(Particle p)
    {
        int i = hash_i(p.pos_x);
        int j = hash_j(p.pos_y);
                
        if ((i>=0) && (i < num_cells) && (j>=0) && (j < num_cells))
            (cell[i][j]).addElement(p);        
    }     
 
    // add the array of particles
    public void add(Particle p[])
    {
        for (int i = 0; i < p.length; i++)
            add(p[i]);      
    }
 
    // given a (x,y) point, return a particle representing the weighted
    // average (based on euclidean distance) of our k closest neighbors
   
    public Particle find_avg_neighbor(float x, float y)
    {
        int i = hash_i(x);
        int j = hash_j(y);
        Particle closest = null;
        Particle part = null;        
        float part_dist, closest_dist;
        int found = 0;
        
        Vector neighbors = new Vector();
        
        // add to our vector all the particle we find in this cell
        
        add_from_cell(i, j, neighbors);
        
        if (neighbors.size() >= average_over)
            found++;
                              
        // continue with bigger rings as long as necessary
        int size = 1;
        while ((found < 2) && (size < num_cells))
        {
            add_from_ring(i, j, size, neighbors);
            if (neighbors.size() >= average_over)
                found++;           
            size++;
        }
                
        return find_average(x, y, neighbors);
    }    
    
    // given a list of Particle, return a weighted average
    Particle find_average(float x, float y, Vector list)
    {
        Particle p = new Particle();
        
        float distance [] = new float[list.size()];
        float weight [] = new float[list.size()];
        int i;
        
        float total_dist = 0.0f;
        
        for (i=0; i<list.size(); i++)
        {
            distance[i] = get_distance(x, y, (Particle) list.elementAt(i));
            total_dist += distance[i];
        }

        // do an inverse weighted average, I made this up myself!
        
        // weight_i = 1 / (dist_i / sum_of_all_dist)
        // then we normalize so the sum of the weights equals 1
        
        float inverse_sum = 0.0f;
        
        for (i = 0; i < list.size(); i++)
        {
            distance[i] = 1.0f / (distance[i] / total_dist);
            inverse_sum += distance[i];    
        }
        
        float weight_check = 0.0f;
        for (i = 0; i < list.size(); i++)
        {
            weight[i] = distance[i] / inverse_sum;
            weight_check += weight[i];            
        }
                                       
        p.density = 0.0f;
        p.pressure = 0.0f;
        p.vel_x = 0.0f;
        p.vel_y = 0.0f;
        
        for (i=0; i<list.size(); i++)        
        {
            p.density += ((Particle) list.elementAt(i)).density * weight[i];
            p.pressure += ((Particle) list.elementAt(i)).pressure * weight[i];
            p.speed += ((Particle) list.elementAt(i)).speed * weight[i];                        
        }
        
        return p;
    }
    
    // add all particles from the cell i,j to the vector list
    void add_from_cell(int i, int j, Vector list)
    {
        if ((i>=0) && (i<num_cells) && (j>=0) && (j<num_cells))
            for (int k=0; k<cell[i][j].size(); k++)
                list.addElement(cell[i][j].elementAt(k));
    }
    
    // add all particle from the cells in a ring around cell (i, j)
    void add_from_ring(int i, int j, int size, Vector list)
    {
        // check the top row of cells        
        for (int k = i - size; k <= i + size; k++)
            add_from_cell(k, j-size, list);
        
        // check the bottom row of cells        
        for (int k = i - size; k <= i + size; k++)
            add_from_cell(k, j+size, list);
            
        // check the left row of cells        
        for (int k = j - size + 1; k <= j + size - 1; k++)
            add_from_cell(i-size, k, list);
            
        // check the right row of cells        
        for (int k = j - size + 1; k <= j + size - 1; k++)
            add_from_cell(i+size, k, list);
    }
    
    // given a (x,y) point, return the attribute field of the closest
    // particle to that point
    
    public Particle find_neighbor(float x, float y)
    {
        int i = hash_i(x);
        int j = hash_j(y);
        Particle closest = null;
        Particle part = null;        
        float part_dist = FLAG;
        float closest_dist = FLAG;
        int found = 0;
        
        // find the closest particle in the current cell
        closest = search_cell(x, y, i, j);        
        closest_dist = get_distance(x, y, closest);     

        // we track how many times we find a candidate
        // we can force a certain search size to eliminate any
        // effect of the square grid missing the nearest neighbor
        if (closest!=null)
            found++;
            
        // we must at the least search the cells surrounding this one,
        // if we still don't have a candidate (all cells empty, we 
        // carry on in bigger and bigger circles)
        
        part = search_ring(x, y, i, j, 1);
        part_dist = get_distance(x, y, part);

        if (part!=null)
            found++;
            
        if (part_dist < closest_dist)
        {
            closest_dist = part_dist;
            closest = part;
        }
        
        // continue with bigger rings if necessary
        int mysize = 2;
        while ((found < 2) && (mysize < num_cells))
        {
            part = search_ring(x, y, i, j, mysize);            
            part_dist = get_distance(x, y, part);
        
            if (part_dist < closest_dist)
            {
                closest_dist = part_dist;
                closest = part;
            }                                     
            
            if (part!=null)
                found++;
                
            mysize++;
        }

        return closest;
    }
    
    // search in a ring of specified size around the current cell
    private Particle search_ring(float x, float y, int i, int j, int sz)
    {
        Particle part = null;
        Particle closest = null;
        float closest_dist = FLAG;
        float part_dist = FLAG;

        // check the top row of cells        
        for (int k = i - sz; k <= i + sz; k++)
        {
            part = search_cell(x, y, k, j-sz);
            part_dist = get_distance(x, y, part);
            
            // check if we've found a new minimum
            if (part_dist < closest_dist)
            {
                closest_dist = part_dist;
                closest = part;
            }
        }

        // check the bottom row of cells        
        for (int k = i - sz; k <= i + sz; k++)
        {
            part = search_cell(x, y, k, j+sz);            
            part_dist = get_distance(x, y, part);
            
            // check if we've found a new minimum
            if (part_dist < closest_dist)
            {
                closest_dist = part_dist;
                closest = part;
            }
        }

        // check the left row of cells        
        for (int k = j - sz + 1; k <= j + sz - 1; k++)
        {
            part = search_cell(x, y, i - sz, k);
            part_dist = get_distance(x, y, part);
            
            // check if we've found a new minimum
            if (part_dist < closest_dist)
            {
                closest_dist = part_dist;
                closest = part;
            }
        }        

        // check the right row of cells        
        for (int k = j - sz + 1; k <= j + sz - 1; k++)
        {
            part = search_cell(x, y, i + sz, k);
            part_dist = get_distance(x, y, part);
            
            // check if we've found a new minimum
            if (part_dist < closest_dist)
            {
                closest_dist = part_dist;
                closest = part;
            }
        }        
        
        return closest;
    }
    
    // search for the closest point to x, y in cell i, j
    // return the minimum distance Particle
    private Particle search_cell(float x, float y, int i, int j)
    {
        float min = FLAG;
        Particle min_part = null;
        float pos_x, pos_y, dist;

        // check to be sure this cell index is in range
        if ((i >= 0) && (i < num_cells) && (j >=0) && (j < num_cells))
        {
            for (int k = 0; k < cell[i][j].size(); k++)
            {
                    pos_x = ((Particle) cell[i][j].elementAt(k)).pos_x;
                    pos_y = ((Particle) cell[i][j].elementAt(k)).pos_y;
            
                    dist = get_distance(pos_x, pos_y, x, y);
            
                    if (dist < min)
                    {
                        min = dist;
                        min_part = (Particle) cell[i][j].elementAt(k);                
                    }
            }
        }
        
        return min_part;        
    }
 
    // compute the Euclidean distance
    private float get_distance(float x1, float y1, float x2, float y2)
    {
        return (float) Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
    }

    // compute the Euclidean distance
    private float get_distance(float x1, float y1, Particle p)
    {
        if (p==null)    
        {
        //    System.out.println("get_distance passed a null particle");
            return FLAG;
        }
        else
        {
            float x2 = p.pos_x;
            float y2 = p.pos_y;
            return (float) Math.sqrt( (x1 - x2)*(x1-x2) + (y1 - y2)*(y1-y2));
        }
    }
    
    // given a x real coordinate, hash to a particular cell
    private int hash_i(float x)
    {
        return (int) (x / (max_x / num_cells));
    }

    // given a y real coordinate, hash to a particular cell
    private int hash_j(float y)
    {
        return (int) (y / (max_y / num_cells));
    }

}