
// Class that handles the socket communication with the server side

import java.io.*;
import java.util.*;
import java.awt.*;

class Network implements globals {
    
    // these are the commands that can be sent to the server
    private final static int BYE       = 0;
    private final static int STATS     = 1;
    private final static int SHUTDOWN  = 2;
    private final static int PARAMS    = 3;
    private final static int WORLD     = 4;
    private final static int NETWORK   = 5;
    private final static int RUNSIM    = 6;

    
    public Client mylink;
    public boolean connected = false;
    public String server = new String();        // string name of our server
    public String clientname = new String();    // string name of our system
    public int port = -1;    
    public int reverse = 1 ;                    // reverse bytes or not    
    public int totalnodes = -1;                 // how many nodes the server actually has
    Vector nodenames = new Vector(32);          // stores the system names of all the servers nodes
    Vector nodeflops = new Vector(32);          // stores the last performance value of each node
    public sphclient myparent;     
    public UpdateReceiver receiver;
    
    public Network(Frame parent)
    {   
        myparent = (sphclient) parent;
        receiver = new UpdateReceiver (myparent);
    }
    
    public boolean connect()
    {
        // startup a socket connection with the current settings
        
        if (connected)
        {
            System.out.println("Already connected!");
            return false;
        }
        
        if ((port==-1) || (server==null) || (server.equals("")))
        {
            System.out.println("Network settings invalid!");
            return false;
        }
        
        try {
            mylink = new Client(port, port+1, server, reverse);
            connected = true;
            clientname = mylink.get_myname();
            
            int[] nums = new int[1];
            
            // now get the total number of nodes our server has
            mylink.recv_ints(nums, 1);
            totalnodes = nums[0];
            double[] flops = new double[totalnodes];
                       
        	for (int i=0; i<totalnodes; i++)
    	    	nodenames.addElement(mylink.recv_string('\n'));           
            
            // receive the names of all the nodes and an initial megaflop score
	        mylink.recv_doubles(flops, totalnodes);
            for (int i=0; i<totalnodes; i++)
    	    	nodeflops.addElement(new Double(flops[i]));
    	
            // send the server our params and world objects
            update_server();
            
            return true;
        } catch (Exception e)
        {
            return false;
        }
    }
    
    public boolean disconnect() 
    {
        try {
            // tell server to say bye
            int[] nums = {BYE};
            mylink.send_ints(nums, 1);
            
            mylink.closesocket();   
            connected = false;
            nodeflops.removeAllElements();
            nodenames.removeAllElements();
            return true;
        } catch (Exception e)
        {
          return false;
        }
    }

    public boolean shutdown() 
    {
        // tell the server to terminate
        
        try {        
            int[] nums = {SHUTDOWN};
            mylink.send_ints(nums, 1);
            
            mylink.closesocket();   
            connected = false;
            nodeflops.removeAllElements();
            nodenames.removeAllElements();
            return true;
        } catch (Exception e)
        {
          return false;
        }
    }

    public boolean update_stats()
    {
        // obtain updated stats from the network
                
       try {
            
            // send the server a benchmark request message
            int[] nums = new int[1];
            nums[0] = STATS;
            mylink.send_ints(nums, 1);

            // now get the total number of nodes our server has
            // (it may have changed)
            mylink.recv_ints(nums, 1);
            totalnodes = nums[0];   
            
            double[] flops = new double[totalnodes];

            nodeflops.removeAllElements();
            nodenames.removeAllElements();                       
        	
        	for (int i=0; i<totalnodes; i++)
    	    	nodenames.addElement(mylink.recv_string('\n'));           
            
            // receive the names of all the nodes and an initial megaflop score
	        mylink.recv_doubles(flops, totalnodes);
            for (int i=0; i<totalnodes; i++)
    	    	nodeflops.addElement(new Double(flops[i]));
    	    	            
            return true;
        } catch (Exception e)
        {
            return false;
        }
    }
    
    public String port_str()
    {
        Integer num = new Integer(port);
        return num.toString();       
    }
    
    // read the network parameters in from a file
    // returns true if successful, false o.w.
    public boolean read_file(String filename)
    {
        String temp;
        
        DataInputStream input;
        
        try {
              input = new DataInputStream(new FileInputStream(filename));
        } 
        catch (IOException e)
            {
                System.out.println("Error opening "+filename);
                return false;
            }
        try {
            port = (new Integer(input.readUTF())).intValue();            
            reverse = (new Integer(input.readUTF())).intValue();
            server = (new String(input.readUTF()));
        }
        catch (Exception e)
        {
            System.out.println("Error reading data from "+filename);
            return false;
        }        

        try {
            input.close();
        }
        catch (IOException e)
        {
            System.out.println("Error closing "+filename);
            return false;        
        }
        
        return true;
    }
    
    // write the network parameters in from a file
    // returns true if successful, false o.w.
    public boolean write_file(String filename)
    {        
        DataOutputStream output;
        
        try {
              output = new DataOutputStream(new FileOutputStream(filename));
        } 
        catch (IOException e)
            {
                System.out.println("Error writing "+filename);
                return false;
            }
        try {
            output.writeUTF((new Integer(port).toString()));    
            output.writeUTF((new Integer(reverse).toString()));
            output.writeUTF(server);
        }
        catch (IOException e)
        {
            System.out.println("Error writing data to "+filename);
            return false;
        }        

        try {
            output.close();
        }
        catch (IOException e)
        {
            System.out.println("Error closing "+filename);
            return false;        
        }
        
        return true;
    }
 
    public boolean send_params()
    {
       // send the server our parameters           
       try {
       
            // get the param objects packet ready to go
            myparent.param.update_packet();
            
            // tell the server that parameters are coming in
            int[] nums = new int[1];
            nums[0] = PARAMS;
            mylink.send_ints(nums, 1);

            mylink.send_doubles(myparent.param.packet, myparent.param.packet.length);
                        
            return true;
        } catch (Exception e)
        {
            return false;
        }
    }

    public boolean send_world()
    {
       // send the server our shapes
       try {
                   
            // tell the server that parameters are coming in
            int[] nums = new int[1];
            nums[0] = WORLD;
            double[] data = new double[5 * myparent.world.num_shapes + 2];
            
            mylink.send_ints(nums, 1);
            
                   
            // tell the server how many shapes to expect
            nums[0] = myparent.world.num_shapes;
            mylink.send_ints(nums, 1);
                   
            // send the packet that the world class will build for us
                
            myparent.world.get_packet(data);
                
            mylink.send_doubles(data, data.length);                
                        
            return true;
        } catch (Exception e)
        {
            return false;
        }
    }

    public boolean run_sim()
    {
       // startup a simulation on the server

       // get rid of any image currently on our animatin canvas
       
       myparent.animation_canvas.img = null;
       
       try {                               
            int[] nums = new int[1];
            nums[0] = RUNSIM;
            
            mylink.send_ints(nums, 1);
            
            // update the server with our param, world 
           
            update_server();

            // make sure we've got the right screen max coordinates
            // in our frame generation classes
          
            receiver.width = myparent.canvas_width;
            receiver.height = myparent.canvas_height;
            
            // fire up our receiver thread
            if (receiver.isAlive())
            {
                receiver.mode = NEWSIM;
                receiver.output_num = 0;    // reset the file counter
                myparent.animation_canvas.cur_frame = 0;
                myparent.animation_canvas.num_frames = 0;
                myparent.update_animation_controls();
                receiver.resume();
            }
            else
            {
                receiver.mode = NEWSIM;
                receiver.start();
            }   
            return true;
            
        } catch (Exception e)
        {
            return false;
        }
    }
        
        
    public boolean update_server()
    {
        // send the server an update of all the parameters and shapes
        // and network settings
        
        if ((send_params()) && (send_world()))
            return true;
        else
            return false;
    }
}