// This is the normal code that does not do lazy updates

#include "SlaveGrid.h"
#include <stdio.h>

//*****************************************************************
//* Constructor
//*****************************************************************
SlaveGrid::SlaveGrid() : Grid()
{
}

//*****************************************************************
//* Another Constructor
//*****************************************************************
SlaveGrid::SlaveGrid(Parameter P) : Grid(P)
{
}
 
//******************************************************************
//  overloaded ostream operator (standard output)
//******************************************************************
ostream & operator << ( ostream & str, const SlaveGrid & G)
{
	for(int i = 0; i < G.endOfPts; i++)
	{
		// only output points that actually really exist
		if (!G.t[i].isEmpty())
			str << G.t[i] << endl;
	}	

	str << "\n";

	for(int j = 0; j < G.numShapePts; j++)
		str << G.shapePtsT[j] << endl;


	return str;
}	

//******************************************************************
// overloaded ofstream operator (file output)
//******************************************************************
ofstream & operator << ( ofstream & fStr, const SlaveGrid & G)
{
	for(int i = 0; i < G.endOfPts; i++)
	{
		// only output points that actually really exist
		if (!G.t[i].isEmpty())
			fStr << G.t[i];
	}	
		

	fStr << "\n";

	for(int j = 0; j < G.numShapePts; j++)
		fStr << G.shapePtsT[j];

	return fStr;
}	
 
 
//*****************************************************************
//* Destructor
//*****************************************************************
SlaveGrid::~SlaveGrid()
{
	// the parent destructor will take care of things
}	


//********************************************************* 
//* This method calculates the pressure of each particle
//********************************************************* 
void SlaveGrid::calcPressure()
{	
	for(int i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())	
		{
			t[i].setMinMaxDensity();	// remember our min and max density
			t[i].calcPressure(param);
		}
	}
	
	for (int i=0; i<numShapePts; i++)
	{
		shapePtsT[i].calcPressure(param);
	}
}


//**********************************************************
//* This method calculates the interactions of each particle
//* in the center cell with other particles in the center cell
//* as well as the interactions of particles in neighboring
//* cells.
//********************************************************* 
void SlaveGrid::calcSPHEqns()
{

	int cellIndexI, cellIndexJ;
	int centerCellFreePtSize;
	int centerCellFreeListLoop;
	int iFree;

	// Zero out the bound particle change in density

	for (int i=0; i<numShapePts; i++)
		shapePtsT[i].zeroSPHVars();
	
	// Loop over all cells in my region of the Grid.

	for(cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
	{
		for(cellIndexJ = startCellY; 
		    cellIndexJ <= endCellY; 
			cellIndexJ++)
		{
			centerCellFreePtSize = 
			         cells[cellIndexI][cellIndexJ].getNumFreePts();
			cells[cellIndexI][cellIndexJ].cur2FirstFreeItem();
	
			//calculates the interactions of each free particle in 
			//cell[cellIndexI][cellIndexJ] with other particles 
			//(free and boundary) in this cell and it's neighbors.
			for(centerCellFreeListLoop = 0; 
		   	    centerCellFreeListLoop < centerCellFreePtSize; 
		        centerCellFreeListLoop++)
			{
				//Get index of current particle in Free list.
				iFree = 
			    cells[cellIndexI][cellIndexJ].getCurrentFreeItem();


				//The ordering of the interactions is done to reduce 
				// the numerical error due introduced when adding
				//a BIG number with a little number. The ordering
				// takes into consideration the smaller contributions
				//first (basically relates to the shape of the kernel)

				if (t[iFree].isEmpty())
					cout << "Panic!  Cell list had an empty particle! " << t[iFree] << endl;

				//Zero out this particle's SPH realtions.
				t[iFree].zeroSPHVars();

				//LEFT-UPPER neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI - 1,
												   cellIndexJ + 1);
				//RIGHT-UPPER neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI + 1,
												   cellIndexJ + 1);
				//RIGHT-LOWER neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI + 1,
												   cellIndexJ - 1);
				//LEFT-LOWER neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI - 1,
												   cellIndexJ - 1);
				// LEFT neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI - 1,
												   cellIndexJ);
				//TOP neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI,
												   cellIndexJ + 1);
				//RIGHT neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI + 1,
												   cellIndexJ);
				//BOTTOM neighbor cell. 
				calcFreeParticleInteraction(iFree, cellIndexI,
												   cellIndexJ - 1);
				//SAME cell. 
				calcFreeParticleInteraction(iFree, cellIndexI, 
				                                   cellIndexJ);

	
				//Now add the contributions from the body forces.
				t[iFree].addForceX(param);
				t[iFree].addForceY(param);
	
				// increment to the next free particle in the 
				// center cell.
				// if the particle got deleted during this loop, 
				// then we've already advanced the pointer

				cells[cellIndexI][cellIndexJ].cur2NextFreeItem();
			} // end free particle loop 
		} // end cellIndexJ loop
	} // end cellIndexI loop

} // end method
		

//*********************************************************************
//* calculates the interactions between the particle given by index iFree
//* And all other free and bnd particles in the cell given by the indeces
//* cellIndexI, cellIndexJ.
//**********************************************************************
void SlaveGrid::calcFreeParticleInteraction(int iFree, int cellIndexI,
                                                  int cellIndexJ)
{
	int jFree, listLoop, size;
	Point v_aB;
	Point v_Ba;
	double oldX;
	double offset = 0.0;

	// Check to make sure we actually have a neighbor in this direction

	if ((cellIndexI > numCellsWide) || (cellIndexI < -1) ||
	   (cellIndexJ < startCellY) || (cellIndexJ > endCellY))
	{
		// don't do anything, this particles neighbor isn't really there
	}	
	else
	{
		//conditionals for periodic boundary conditions.
		if(cellIndexI == -1)
		{

			cellIndexI = numCellsWide - 1;
			offset = -1.0 * numCellsWide * cellSize;
		}
		else if(cellIndexI == numCellsWide)
		{
			cellIndexI = 0;
			offset = 1.0 * numCellsWide * cellSize;
		}
		else
			offset = 0.0; 


		// We first loop over the free particles, and then loop 
		// over bnd.
		// Point to first particle of free particle list.
		cells[cellIndexI][cellIndexJ].other2FirstFreeItem();
		
		//Get this list's size
		size = cells[cellIndexI][cellIndexJ].getNumFreePts();

		//calculate the interactions of particle iFree with other 
		//free particles from this cell. 

		for(listLoop = 0; listLoop < size;  listLoop++)
		{
			jFree = cells[cellIndexI][cellIndexJ].getOtherFreeItem();	
			if((iFree != jFree) && (!t[jFree].isEmpty()))
			{	
				v_aB.setXY(
				 (t[iFree].getVelocityX() - t[jFree].getVelocityX()),
				 (t[iFree].getVelocityY() - t[jFree].getVelocityY()));
	
				oldX = t[jFree].getPositionX();
				t[jFree].setPositionX(oldX + offset);

				if (TRACK_NEIGHBORS)
					t[iFree].updateNumNeighbors(t[jFree]);

				t[iFree].calcChangeDensity(t[jFree], v_aB);
				t[iFree].calcChangeVelocity(t[jFree], v_aB, param);			
				t[jFree].setPositionX(oldX);

			}
				
			//increment to the next item in the free list.
			cells[cellIndexI][cellIndexJ].other2NextFreeItem();
		}

		int jBnd;
	
		// Now loop over bnd points.
		//Point to first particle of bnd particle list.
		cells[cellIndexI][cellIndexJ].other2FirstBndItem();
		
		//Get this list's size
		size = cells[cellIndexI][cellIndexJ].getNumBndPts();

		//calculate the interactions of particle iFree with other 
		//bnd particles from this cell. 
		for(listLoop = 0; listLoop < size;  listLoop++)
		{
			jBnd = cells[cellIndexI][cellIndexJ].getOtherBndItem();	


			oldX = shapePtsT[jBnd].getPositionX();
			shapePtsT[jBnd].setPositionX(oldX + offset);

			calcVaB(v_aB, t[iFree], shapePtsT[jBnd]);
			// This is the mirror interaction of the boundary
			//particle on this free particle.
			v_Ba.setXY(-1.0*v_aB.getX(), -1.0*v_aB.getY()); 

			if (TRACK_NEIGHBORS)
				t[iFree].updateNumNeighbors(shapePtsT[jBnd]);

			shapePtsT[jBnd].calcChangeDensity(t[iFree],  v_Ba);
			t[iFree].calcChangeDensity(shapePtsT[jBnd], v_aB);
			t[iFree].calcChangeVelocity(shapePtsT[jBnd], v_aB, param);

			shapePtsT[jBnd].setPositionX(oldX );

			//increment to the next item in the bnd list.
			cells[cellIndexI][cellIndexJ].other2NextBndItem();
		}
	} // end neighbor else
}	


//*********************************************************************
//* calculates the interactions between the bnd particle given by index 
//* iBnd and all other free particles in the cell given by the indeces
//* cellIndexI, cellIndexJ.
//**********************************************************************
void SlaveGrid::calcBndParticleInteraction(int iBnd, int cellIndexI, 
 						int cellIndexJ)
{
	int jFree, listLoop, size;
	Point v_aB(0.0, 0.0);  
	Point v_Ba(0.0, 0.0);  // calcVaB() expects the arguements
	                       // in a certain order, but this order
						   // doesn't work for boundary interaction
						   // so we must reverse the 'vector' on
						   // this side of the function call.
	
	//Point to first particle of boundary particle list.
	cells[cellIndexI][cellIndexJ].other2FirstFreeItem();
		
	//Get the list's size.
	size = cells[cellIndexI][cellIndexJ].getNumFreePts();

	//calculate the interactions of particle iBnd with other 
	//free particles from this cell. 
	for(listLoop = 0; listLoop < size;  listLoop++)
	{
		jFree = cells[cellIndexI][cellIndexJ].getOtherFreeItem();	
		if (!t[jFree].isEmpty())
		{
			calcVaB(v_aB, t[jFree], shapePtsT[iBnd]);
			v_Ba.setXY(-1.0*v_aB.getX(), -1.0*v_aB.getY()); 
			shapePtsT[iBnd].calcChangeDensity(t[jFree],  v_Ba);
		
			//increment to the next item in the boundary list.
			cells[cellIndexI][cellIndexJ].other2NextFreeItem();
		}
	}


}	





//***********************************************************************
//*This method is used in the leap frog calculation and is used to 
//*predict values to be used to calculate more accurate results.
//**********************************************************************
void SlaveGrid::predictParticles()
{
	int i;
	double dt_over_2 = param.deltaT / 2.0;
	
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{	
			t[i].setVelocityX(tMinusOneHalf[i].getVelocityX() + 
				dt_over_2 * t[i].getChangeVelocityX()); 
			t[i].setVelocityY(tMinusOneHalf[i].getVelocityY() + 
				dt_over_2 * t[i].getChangeVelocityY()); 
			t[i].setDensity(tMinusOneHalf[i].getDensity()     + 
				dt_over_2 * t[i].getChangeDensity()); 
		}
	}

	for(i = 0; i < numShapePts; i++)
	{
		shapePtsT[i].setDensity(shapePtsTMinusOneHalf[i].getDensity() + 
			dt_over_2 * shapePtsT[i].getChangeDensity()); 
	}	
}



//***********************************************************************
//* This method is a special case. It is only done during the first
//* loop of the leap frog method and the only difference between it
//* and a normal update is the time step and where previous time step
//* values come from. For the t = 0 case, they come from the intial
//* conditions, or the t[] array.
//*********************************************************************
void SlaveGrid::updateParticlesFirstLoop()
{
	double dt_over_2 = param.deltaT / 2.0;
	int i;

	
	//update each free particle's position, velocity, and density.
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{

		tPlusOneHalf[i].setVelocityX( 
		    t[i].getVelocityX() + 
			dt_over_2 * t[i].getChangeVelocityX()); 
		tPlusOneHalf[i].setVelocityY(
		    t[i].getVelocityY() + 
			dt_over_2 * t[i].getChangeVelocityY()); 
		tPlusOneHalf[i].setPositionX(t[i].getPositionX() + 
			param.deltaT * (tPlusOneHalf[i].getVelocityX()));
		tPlusOneHalf[i].setPositionY(t[i].getPositionY() + 
		 	param.deltaT * (tPlusOneHalf[i].getVelocityY()));
		tPlusOneHalf[i].setDensity(t[i].getDensity() + 
		            dt_over_2 * t[i].getChangeDensity()); 

		}//end if
	}//end for

	//Now move all particles to the correct time step array.
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{
			tMinusOneHalf[i] = tPlusOneHalf[i];
			t[i].setPositionX(tPlusOneHalf[i].getPositionX());
			t[i].setPositionY(tPlusOneHalf[i].getPositionY());
		}
	}

	for(i = 0; i < numShapePts; i++)
	{
		shapePtsTPlusOneHalf[i].setDensity(shapePtsT[i].getDensity() + 
		            dt_over_2 * shapePtsT[i].getChangeDensity()); 

	}//end for

	//Now move all shape particles to the correct time step array.
	for(i = 0; i < numShapePts; i++)
		shapePtsTMinusOneHalf[i] = shapePtsTPlusOneHalf[i];

}		



//*********************************************************************
//* This method will be used for all loops except the first. It updates
//* the particles for the full time step (thus the use of dt and not 
//* dt/2.0 and also the use of tMinusOneHalf[] array).
//* It caclulates the new Velocity, position, and density from the old
//* values (full time step prior) and also uses the current changes in
//* density and velocity to aid in this process.
//********************************************************************* 
void SlaveGrid::updateParticles()
{
	int i;
	
	//update each free particle's position, velocity, and density.
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{
			tPlusOneHalf[i].setVelocityX(
			    tMinusOneHalf[i].getVelocityX() + 
				param.deltaT * t[i].getChangeVelocityX()); 
			tPlusOneHalf[i].setVelocityY(
			    tMinusOneHalf[i].getVelocityY() + 
				param.deltaT * t[i].getChangeVelocityY()); 
			tPlusOneHalf[i].setPositionX(
			    tMinusOneHalf[i].getPositionX() + 
				param.deltaT * (tPlusOneHalf[i].getVelocityX()));
			tPlusOneHalf[i].setPositionY(
			    tMinusOneHalf[i].getPositionY() + 
				param.deltaT * (tPlusOneHalf[i].getVelocityY()));
			tPlusOneHalf[i].setDensity(tMinusOneHalf[i].getDensity() + 
	            		param.deltaT * t[i].getChangeDensity()); 
		}
	}

	//Now move all particles to the correct time step array.
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{
			tMinusOneHalf[i] = tPlusOneHalf[i];
			t[i].setPositionX(tPlusOneHalf[i].getPositionX());
			t[i].setPositionY(tPlusOneHalf[i].getPositionY());
		}	
	}


	for(i = 0; i < numShapePts; i++)
		shapePtsTPlusOneHalf[i].setDensity(shapePtsTMinusOneHalf[i].getDensity() + 
		            param.deltaT * shapePtsT[i].getChangeDensity()); 

	//Now move all shape particles to the correct time step array.
	for(i = 0; i < numShapePts; i++)
		shapePtsTMinusOneHalf[i] = shapePtsTPlusOneHalf[i];


}		

//********************************************************************
// This is used by the slave to set its SlaveGrid size
//********************************************************************
void SlaveGrid::setRangeOfRows(int start, int end)
{
        startCellY = start;
        endCellY = end;
}



//********************************************************************
//* This method checks all cells to see if any particles need to be 
//* moved to another cell. If it does, then it moves them.
//*********************************************************************
void SlaveGrid::updateGrid()
{
	int size;
	int i, cellIndexI, cellIndexJ, loop;
	int newI, newJ;

	//loop over all cells in SlaveGrid. For each free particle in each cell, 
	//check to see if it has left the current cell and if it has, then 
	//remove it's index from that cell and hash it to the correct cell.
	// We go backwards because the flow moves left to right and we 
	//stand the chance of checking every particle twice. Once when we
	// move it to the right cell and then again when we are updating
	// the right cell to see if it belongs there.


	int start, end;
	
	if (southMost)
		start = startCellY;
	else
		start = startCellY + 1;

	if (northMost)
		end = endCellY;
	else
		end = endCellY + 1;

	for(cellIndexI = numCellsWide - 1; cellIndexI >=0; cellIndexI--)
	{
	   	for(cellIndexJ = start; cellIndexJ <= end; cellIndexJ++)
		{
			//Point to first free particle in this cell.
			cells[cellIndexI][cellIndexJ].cur2FirstFreeItem();
		
			//Get the number of free particles in this cell and then loop
			//over all of them, moving them if necessary.
			size = cells[cellIndexI][cellIndexJ].getNumFreePts();
			for(loop = 0; loop < size; loop++)
			{
				i = cells[cellIndexI][cellIndexJ].getCurrentFreeItem();
			
				newI = hashI(t[i].getPositionX()); 
				newJ = hashJ(t[i].getPositionY()); 

				if (t[i].isEmpty())
				{
					cout << debugTimeSteps << "  Panic!  Empty particle in our cell LL." << endl;
					cout << t[i] << endl;
				}
			
			
				// Let's make sure that this particle hasn't flown
				// off of the SlaveGrid (i.e. canvas).
				if ((newI<0) || (newI>=numCellsWide) ||
				    (newJ<0) || (newJ>=numCellsHigh) )
				{
					// simulate periodic boundary conditions.
					if(newI == numCellsWide)
					{
						double offset = numCellsWide*cellSize;

						t[i].setPositionX(t[i].getPositionX() - offset);
						tPlusOneHalf[i].setPositionX(tPlusOneHalf[i].getPositionX()
						                              - offset);
						tMinusOneHalf[i].setPositionX(tMinusOneHalf[i].getPositionX() 
						                              - offset);
						cells[0][newJ].pushFree(i);
						cells[cellIndexI][cellIndexJ].removeCurrentFreeItem();
					}
					else
					{
						deletePt(i);
						cells[cellIndexI][cellIndexJ].removeCurrentFreeItem();
					}
				}

				// Use the hashing function to see if the particle 
				// has moved.
				else if(newJ != cellIndexJ)
				{
					// the row number has changed

					// see if the particle has moved
					// into the north buffer row or
					// above, these will need to be
					// sent to our neighbor

					if (newJ <= startCellY) 
					{
						// if it has moved to our
						// buffer row, add it to
						// the cell because it
						// won't be sent from neighbor yet
						if (newJ == startCellY)
							cells[newI][newJ].pushFree(i);					

						// we only want to add it to the south bound list if we aren't the bottom processor
						if (!southMost)
							southBoundPts.push(i);
		
						// delete this points for the cell LL
						// the actual array entry for it will be 
						// around until we delete the
						// south and north bound LLs
						cells[cellIndexI][cellIndexJ].removeCurrentFreeItem();
					}
					// similarly with southbound particles
					else if (newJ >= endCellY)
					{
						if (newJ == endCellY)
							cells[newI][newJ].pushFree(i);					

						if (!northMost)
							northBoundPts.push(i);

						cells[cellIndexI][cellIndexJ].removeCurrentFreeItem();
					}
					else
					{
						// this guy just moved within our normal cells
						cells[newI][newJ].pushFree(i);
						cells[cellIndexI]
					   		 [cellIndexJ].removeCurrentFreeItem();
					}
				}
				else if(newI != cellIndexI)
				{
					cells[newI][newJ].pushFree(i);
					cells[cellIndexI][cellIndexJ].removeCurrentFreeItem();
				}

				else
				{	//if the particle was deleted, then the pointer 
					//automatically points to the next node in the list.
					//thus, increasing is not necessary if the 'if' 
					//condition is true.
					cells[cellIndexI][cellIndexJ].cur2NextFreeItem();
				} // end if
			} // end cell loop
		} // end row loop
	} // end column loop
} // end updateGrid




//*******************************************************************
// free up the array space taken by points in our north most and south
// most cells
//*******************************************************************
void SlaveGrid::deleteBufferRows()
{
	if (!southMost)
		deleteRow(startCellY);

	if (!northMost)
		deleteRow(endCellY);
}

void SlaveGrid::deleteRow(int cellIndexJ)
{
	for(int cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
	{
		//Point to first free particle in this cell.
		cells[cellIndexI][cellIndexJ].cur2FirstFreeItem();
			
		//Get the number of free particles in this cell and then loop
		//over all of them, moving them if necessary.
		int size = cells[cellIndexI][cellIndexJ].getNumFreePts();

		for(int loop = 0; loop < size; loop++)
		{
			int i = cells[cellIndexI][cellIndexJ].getCurrentFreeItem();

			deletePt(i);
			cells[cellIndexI][cellIndexJ].cur2NextFreeItem();
		}

		// now free up the linked list associated with this cell
		cells[cellIndexI][cellIndexJ].deleteFree();
	}
}


//*******************************************************************
// Free up the array spots for everything in the northbound list
// this should be done after we have communicated the points, run
// another time step, but before we put more in the northbound list
// (after a time step, the processor north should be keeping track
//  of these particles)
//*******************************************************************
void SlaveGrid::deleteNorthBound()
{
	northBoundPts.firstCur();

	while (northBoundPts.getCur()!=-1)
	{
		// delete the point from our array

		// only delete this point if it hasn't already
		// been handled by deleteBufferRows (in almost
		// all cases it should be, unless the particle has
		// jumped past the buffer row).
		if (!t[northBoundPts.getCur()].isEmpty())
			deletePt(northBoundPts.getCur());

		// free up the node in the LL
		northBoundPts.deleteCur();	
	}
}

//*******************************************************************
//*******************************************************************
void SlaveGrid::deleteSouthBound()
{
	southBoundPts.firstCur();

	while (southBoundPts.getCur()!=-1)
	{
		// delete the point from our array

		// only delete this point if it hasn't already
		// been handled by deleteBufferRows (in almost
		// all cases it should be, unless the particle has
		// jumped past the buffer row).

		if (!t[southBoundPts.getCur()].isEmpty())
			deletePt(southBoundPts.getCur());

		// free up the node in the LL
		southBoundPts.deleteCur();	
	}
}


//*******************************************************************
// Puts the indices of all north bound migrants into an array
// Allocates memory for the array, returns the number of elements
// that are in that array
//
// We receive the address of the pointer so we can allocate the memory
// have have it passed back 
//*******************************************************************
int SlaveGrid::getNorthBound(int **pts)
{
	// figure out how many migrants there are

	northBoundPts.firstCur();
	
	int len = northBoundPts.length();

	if (len > 0)
	{	
		*pts = new int[len];

		for (int i=0; i<len; i++)
		{
			(*pts)[i] = northBoundPts.getCur();
			northBoundPts.nextCur();
		}

	}

	return len;
}

//*******************************************************************
//*******************************************************************
int SlaveGrid::getSouthBound(int **pts)
{
	// figure out how many migrants there are

	southBoundPts.firstCur();
	
	int len = southBoundPts.length();

	if (len > 0)
	{

		*pts = new int[len];

		for (int i=0; i<len; i++)
		{
			(*pts)[i] = southBoundPts.getCur();

			southBoundPts.nextCur();
		}

	}

	return len;
}

//*******************************************************************
// Get the indices of all particles in our fringe row
//*******************************************************************
int SlaveGrid::getSouthFringe(int **pts)
{
	int len = 0;
	int cellIndexI;
	int index;

	// figure out how many we have in the fringe row

	for(cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
		len += cells[cellIndexI][startCellY+1].getNumFreePts();

	if (len > 0)
	{

		*pts = new int[len];

		int fringeCol = 0;

		// make sure the cells are starting at the beginning
		cells[0][startCellY+1].cur2FirstFreeItem();

		for (int k=0; k<len; k++)
		{
			index = cells[fringeCol][startCellY+1].getCurrentFreeItem();

			// see if we need to advance to the next column
			while (index==-1)
			{
				fringeCol++;
				if (fringeCol >= numCellsWide)
				{
					cout << "Panic!  We didn't find any more particles in north fringe!" << endl;
					return -1;
				}

				// reset to beginning of cell LL		
				cells[fringeCol][startCellY+1].cur2FirstFreeItem();
				index = cells[fringeCol][startCellY+1].getCurrentFreeItem();
			}	

			(*pts)[k] = index;
			cells[fringeCol][startCellY+1].cur2NextFreeItem();
		}
	}

	return len;
}

//*******************************************************************
// Get the indices of all particles in our fringe row
//*******************************************************************
int SlaveGrid::getNorthFringe(int **pts)
{
	int len = 0;
	int cellIndexI;
	int index;

	// figure out how many we have in the fringe row

	for(cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
		len += cells[cellIndexI][endCellY-1].getNumFreePts();

	if (len > 0)
	{

		*pts = new int[len];  
  
	    	int fringeCol = 0;
    
	    	// make sure the cells are starting at the beginning
    		cells[0][endCellY-1].cur2FirstFreeItem();
    
	    	for (int k=0; k<len; k++)
    		{
	    		index = cells[fringeCol][endCellY-1].getCurrentFreeItem();
    
    			// see if we need to advance to the next column
	    		while (index==-1)
	    		{
	    			fringeCol++;
    				if (fringeCol >= numCellsWide)
    				{
    					cout << "Panic!  We didn't find any more particles in north fringe!" << endl;
    					return -1;
	    			}	
    
    				// reset to beginning of cell LL		
	    			cells[fringeCol][endCellY-1].cur2FirstFreeItem();
    				index = cells[fringeCol][endCellY-1].getCurrentFreeItem();
	    		}	
    
    			(*pts)[k] = index;
    
    			cells[fringeCol][endCellY-1].cur2NextFreeItem();
	    	}
	}

	return len;
}


//*******************************************************************
// Return all the indices of all our body particles
//*******************************************************************
int SlaveGrid::getBody(int **pts)
{
	int len = 0;
	int cellIndexI, cellIndexJ, index;

	// find out how many free particles are in the body of our SlaveGrid
	// this excludes the north and southmost rows
	
	int start, end;

	start = startCellY;
	end = endCellY;

	if (!northMost)
		end--;
	if (!southMost)
		start++;	

	for(cellIndexJ = start; cellIndexJ <= end; cellIndexJ++)
		for(cellIndexI = 0; cellIndexI < numCellsWide; cellIndexI++)
			len += cells[cellIndexI][cellIndexJ].getNumFreePts();

	if (len > 0)
	{
		int bodyRow = start;
		int bodyCol = 0;

		*pts = new int[len];

		// make sure we are starting at the beginning of the first cell
		cells[0][start].cur2FirstFreeItem();


		for (int k=0; k<len; k++)
		{
			index = cells[bodyCol][bodyRow].getCurrentFreeItem();

			// see if we have reached the end of this cell, if so
			// then go on to the next one
			while (index == -1)
			{
				bodyCol++;
	
				if (bodyCol >= numCellsWide)
				{
					// we have reached the end of a row, go to the next one
		
					bodyCol = 0;
					bodyRow++;
	
					// see if we've been asked to find more body particles then
					// actually exist
					if (bodyRow > end) 
					{
						cout << "Panic!  We didn't find any more particles in the body!" << endl;
						return -1;
					}
				}
		
				// reset to beginning of cell LL		
				cells[bodyCol][bodyRow].cur2FirstFreeItem();
				index = cells[bodyCol][bodyRow].getCurrentFreeItem();
			}

			(*pts)[k] = index;
			cells[bodyCol][bodyRow].cur2NextFreeItem();
		}
	}

	return len;
}

//*******************************************************************
// Calculates the velocity difference between a free point and a bound
// point.
//*******************************************************************
void SlaveGrid::calcVaB(Point & v_aB, SmoothParticle & sp_a, SmoothParticle & sp_B)
{
	// First, make sure that sp_B is a bound point.
	if (sp_B.getShapeID() >= 0)
	{
		double db_over_da = 
		    theShapes[sp_B.getShapeID()]->calcRatioOfDistances(
				sp_a.getPosition(), sp_B.getPosition());

		if (1+db_over_da < param.betaMax)
		{
			v_aB.setXY((1+db_over_da)*
				sp_a.getVelocityX(), 
				(1+db_over_da)*sp_a.getVelocityY());
		}
		else
		{
			v_aB.setXY(param.betaMax*sp_a.getVelocityX(), 
				param.betaMax*sp_a.getVelocityY());
		}
	}		

}

//*******************************************************************
// Calculates the correct timestep given the current setup of the
// system. This includes such things as forces acting on all
// particles.
//*******************************************************************
void SlaveGrid::calcCorrectDeltaT()
{
	int i;
	double accelX, accelY;
	double maxAccelSquared;
	double accelSquared;
	
	maxAccelSquared = 0;
	
	for(i = 0; i < endOfPts; i++)
	{
		if (!t[i].isEmpty())
		{
			accelX = t[i].getChangeVelocityX();
			accelY = t[i].getChangeVelocityY();
			accelSquared = accelX*accelX + accelY*accelY;
			if (accelSquared > maxAccelSquared)
				maxAccelSquared = accelSquared;
		}	
	}


	double calcDeltaT = 10000000;
	if(maxAccelSquared > 0.0)
		calcDeltaT = 0.25*sqrt(param.h / sqrt(maxAccelSquared));
	
	if(calcDeltaT < param.deltaT)
	{
		cout << "current DT: "<< param.deltaT << " is not small enough.\n";
		cout << "It sjould be: " << calcDeltaT << endl;
	}
}

void SlaveGrid::switchBound()
{
}

